import os
import subprocess
import yaml

from base64 import b64decode

from charms.reactive import (
    set_state,
    remove_state,
    when,
    when_not,
    when_any,
)

from charmhelpers.core import (
    hookenv,
    host,
)

from charms.layer import snap

import shutil

SNAP_DATA = '/var/snap/prometheus-mongodb-exporter/current/'
CA_PATH = os.path.join(SNAP_DATA, 'cacert.pem')
CERT_PATH = os.path.join(SNAP_DATA, 'server.pem')

ARGS_PATH = '/var/snap/prometheus-mongodb-exporter/current/daemon_arguments'
SYSTEMD_SERVICE = 'snap.prometheus-mongodb-exporter.mongodb-exporter.service'

JUJU_ROOT = '/var/lib/juju'

MIN_PORT = 1
MAX_PORT = 65535


@when_not('prometheus-mongodb-exporter.installed')
def install_prometheus_mongodb_exporter():
    ''' Install the proxy application '''
    ca = hookenv.config()['mongodb_tls_ca']
    cert = hookenv.config()['mongodb_tls_cert']
    uri = hookenv.config()['mongodb_uri']

    snap.install('prometheus-mongodb-exporter')

    if not uri:
        message = 'mongodb_uri must be set'
        hookenv.log(message)
        hookenv.status_set('blocked', message)
        return

    # Create a systemd unit file, and open the port.
    port = hookenv.config()['port']
    if not sensible_port(port):
        message = 'port must be between {} and {}'.format(MIN_PORT, MAX_PORT)
        hookenv.log(message)
        hookenv.status_set('blocked', message)
        return

    write_daemon_args(port, uri, ca, cert)
    write_certificates(ca, cert)
    hookenv.open_port(port)
    hookenv.status_set('active', 'serving metrics')
    set_state('prometheus-mongodb-exporter.installed')


def sensible_port(port):
    return (port < MAX_PORT and port > MIN_PORT)


@when('config.changed.port')
def update_opened_port():
    config = hookenv.config()

    new_port = config['port']
    old_port = config.previous('port')

    if old_port and sensible_port(old_port):
        hookenv.close_port(old_port)

    if not sensible_port(new_port):
        message = 'port must be between {} and {}'.format(MIN_PORT, MAX_PORT)
        hookenv.log(message)
        hookenv.status_set('blocked', message)
        return

    hookenv.open_port(new_port)


@when('prometheus-mongodb-exporter.installed')
@when_any('config.changed.port', 'config.changed.mongodb_uri',
          'config.changed.mongodb_tls_hostname_verfication',
          'config.changed.mongodb_tls_ca', 'config.changed.mongodb_tls_cert')
def update_daemon_args():
    config = hookenv.config()
    port = config['port']
    uri = config['mongodb_uri']
    ca = config['mongodb_tls_ca']
    cert = config['mongodb_tls_cert']

    if not sensible_port(port):
        message = 'port must be between {} and {}'.format(MIN_PORT, MAX_PORT)
        hookenv.log(message)
        hookenv.status_set('blocked', message)
        return

    write_daemon_args(port, uri, ca, cert)
    subprocess.check_call(['systemctl', 'daemon-reload'])
    set_state('prometheus-mongodb-exporter.do-restart')


def write_daemon_args(port, mongodb_uri, ca, cert):
    listen_addr = ':{}'.format(port)
    args_dict = {'web.listen-address': listen_addr,
                 'mongodb.uri': mongodb_uri}

    # Handle the CA
    if ca:
        args_dict['mongodb.tls-ca'] = CA_PATH

    # Handle the certificate
    if cert:
        args_dict['mongodb.tls-cert'] = CERT_PATH

    args = ''
    for k, v in args_dict.items():
        args = '{} -{} {}'.format(args, k, v)

    args = 'ARGS="{}"'.format(args)

    with open(ARGS_PATH, 'w') as f:
        f.write(args)

    set_state('prometheus-mongodb-exporter.do-restart')


def write_certificates(ca, cert):
    write_ca(ca)
    write_client_cert(cert)


def write_ca(ca):
    if ca:
        if ca == 'juju':
            machine_id = os.environ['JUJU_MACHINE_ID']
            machine_str = "machine-{}".format(machine_id)
            agentconfig_path = os.path.join(JUJU_ROOT, 'agents',
                                            machine_str, 'agent.conf')

            ca_cleartext = yaml.safe_load(open(agentconfig_path))['cacert']
            ca_cleartext = ca_cleartext.encode('UTF-8')
        else:
            ca_cleartext = b64decode(ca)

        host.write_file(CA_PATH, ca_cleartext, perms=0o400)
    else:
        try:
            os.remove(CA_PATH)
        except FileNotFoundError:
            pass


def write_client_cert(cert):
    if cert:
        if cert == 'juju':
            shutil.copy(os.path.join(JUJU_ROOT, 'server.pem'), CERT_PATH)
        else:
            host.write_file(CERT_PATH, b64decode(cert), perms=0o400)
    else:
        try:
            os.remove(CERT_PATH)
        except FileNotFoundError:
            pass


@when('prometheus-mongodb-exporter.installed')
@when_any('config.changed.mongodb_tls_ca', 'config.changed.mongodb_tls_cert')
def update_certificates():
    write_certificates(hookenv.config()['mongodb_tls_ca'],
                       hookenv.config()['mongodb_tls_cert'])

    set_state('prometheus-mongodb-exporter.do-restart')


@when('prometheus-mongodb-exporter.do-restart')
def do_restart():
    hookenv.log('Restarting prometheus-mongodb-exporter')
    hookenv.status_set('active', 'serving metrics')
    subprocess.check_call(['systemctl', 'restart',
                           SYSTEMD_SERVICE])
    remove_state('prometheus-mongodb-exporter.do-restart')
